# from aiohttp import ClientSession
import os.path
from collections import Counter, defaultdict
from glob import glob
from typing import Any
from urllib.parse import urlparse

from art_dl.cache import cache
from art_dl.sites.deviantart.common import SLUG, make_cache_key
from art_dl.utils.download import download_binary
from art_dl.utils.path import mkdir
from art_dl.utils.print import counter2str
from art_dl.utils.proxy import ClientSession, ProxyClientSession

from .common import logger, progress
from .service import DAService


def parse_link(url: str) -> dict[str, str]:
	parsed = urlparse(url)
	path = parsed.path.lstrip('/').split('/')
	artist = path[0]

	if len(path) == 1 or (len(path) > 2 and path[2] == 'all'):
		# https://www.deviantart.com/<artist>
		# https://www.deviantart.com/<artist>/gallery/all
		return {
			'type': 'all',
			'artist': artist
		}

	if len(path) == 2 and path[1] == 'gallery':
		# https://www.deviantart.com/<artist>/gallery
		# it's "Featured" collection
		return {
			'type': 'folder',
			'folder': 'featured',
			'artist': artist
		}

	if path[1] == 'gallery':
		# https://www.deviantart.com/<artist>/gallery/<some number>/<gallery name>
		# gallery name in format one-two-etc
		return {
			'type': 'folder',
			'folder': path[3],
			'artist': artist
		}

	if path[1] == 'art':
		# https://www.deviantart.com/<artist>/art/<name>
		return {
			'type': 'art',
			'url': url,
			'artist': artist,
			'name': path[2]
		}

	return {
		'type': 'unknown',
		'artist': artist
	}


# download images


async def save_from_url(session: ClientSession, url: str, folder: str, name: str):
	ext = os.path.splitext(urlparse(url).path)[1]
	filename = os.path.join(folder, name + ext)
	if os.path.exists(filename):
		return logger.info('skip existing file', name, progress=progress)

	logger.info('download file', name, progress=progress)
	await download_binary(session, url, filename)


async def save_art(service: DAService, session: ClientSession, art: Any, folder: str):
	url: str = art['url']
	name = url.rsplit('/', 1)[-1]

	if (premium_folder_data := art.get('premium_folder_data')) is not None:
		if premium_folder_data['has_access'] is False:
			logger.warn('no access to', name + ',', 'downloading preview', progress=progress)

	if art['is_downloadable'] is False or art['download_filesize'] == art['content']['filesize']:
		return await save_from_url(session, art['content']['src'], folder, name)

	original_url = await service.get_download(art['deviationid'])
	if original_url is not None:
		await save_from_url(session, original_url, folder, name)


# wrappers for common actions


async def download_folder_by_id(service: DAService, save_folder: str, artist: str, folder_id: str):
	# this session for downloading images
	async with ProxyClientSession() as session:
		async for art in service.list_folder_arts(artist, folder_id):
			await save_art(service, session, art, save_folder)


async def download_art_by_id(service: DAService, deviationid: str, folder: str):
	art = await service.get_art_info(deviationid)
	async with ProxyClientSession() as session:
		await save_art(service, session, art, folder)


# helpers


def is_art_exists(folder: str, artist: str, name: str):
	return len(glob(f'{folder}/{artist}/{name}.*')) > 0


# main functions


async def download(urls: list[str], data_folder: str):
	stats = Counter()  # type: ignore
	progress.total = len(urls)

	service = DAService()

	# ['artist1', ...]
	mapping_all: list[str] = []
	# { '<artist>': ['folder1', ...] }
	mapping_folder: dict[str, list[str]] = defaultdict(list)
	# { '<artist>': [{ 'name': 'name1', 'url': 'url1' }, ...] }
	mapping_art: dict[str, list[dict[str, str]]] = defaultdict(list)

	# group urls by types and artists
	for u in urls:
		parsed = parse_link(u)
		t = parsed['type']
		a = parsed['artist']
		if t == 'all':
			mapping_all.append(a)
		elif t == 'folder':
			mapping_folder[a].append(parsed['folder'])
		elif t == 'art':
			n = parsed['name']
			if is_art_exists(data_folder, a, n):
				stats.update(skip=1)
				progress.i += 1

				logger.info('skip existing', a + '/' + n, progress=progress)
				continue

			deviationid = cache.select(SLUG, make_cache_key(a, u))
			if deviationid is not None:
				stats.update(download=1)
				progress.i += 1

				logger.info('download cached', a + '/' + n, progress=progress)
				save_folder = os.path.join(data_folder, a)
				mkdir(save_folder)
				await download_art_by_id(service, deviationid, save_folder)
				continue

			mapping_art[a].append({
				'name': n,
				'url': u
			})
		elif t == 'unknown':
			stats.update(skip=1)
			progress.i += 1

			logger.warn('unsupported link', u, progress=progress)
			continue

	# process

	# save artists all arts
	for artist in mapping_all:
		stats.update(download=1)
		progress.i += 1

		save_folder = os.path.join(data_folder, artist)
		mkdir(save_folder)
		logger.info('artist', artist, progress=progress)

		await download_folder_by_id(service, save_folder, artist, 'all')

	# save collections
	for artist, folder_list in mapping_folder.items():
		stats.update(download=1)
		progress.i += 1

		save_folder = os.path.join(data_folder, artist)
		mkdir(save_folder)
		logger.info('artist', artist, progress=progress)

		async for folder in service.list_folders(artist):
			if folder['name'] in folder_list:
				logger.info('gallery', folder['pretty_name'], progress=progress)
				await download_folder_by_id(service, save_folder, artist, folder['id'])

	# save single arts
	async with ProxyClientSession() as session:
		for artist, art_list in mapping_art.items():
			progress.i += 1

			all_urls = set(map(lambda a: a['url'], art_list))

			save_folder = os.path.join(data_folder, artist)
			mkdir(save_folder)
			logger.info('artist', artist, progress=progress)

			async for art in service.list_folder_arts(artist, 'all'):
				url = art['url']
				if any(filter(lambda a: a['url'] == url, art_list)):  # type: ignore
					await save_art(service, session, art, save_folder)

					all_urls.remove(url)
					if len(all_urls) == 0:
						break

			if len(all_urls) > 0:
				stats.update(partial_download=1)
				logger.warn('not found', len(all_urls), f'arts ({artist}):', progress=progress)
				for u in all_urls:
					logger.warn(' ', u)
			else:
				stats.update(download=1)

	logger.info(counter2str(stats))
	logger.newline(normal=True)



def parse_range(rng: str) -> list[int] | None:
	"""
	Convert range like `1-3,5` to `[1, 2, 3, 5]`.
	Result is sorted.
	Empty string converts to `None`
	"""
	if rng == '':
		return None

	parts = rng.split(',')
	result: set[int] = set()
	for p in parts:
		if '-' in p:
			[start, end] = p.split('-')
			result.update(range(int(start), int(end) + 1))
		else:
			result.add(int(p))

	return sorted(list(result))



from typing import Any, Callable, Coroutine

# import for mypy
from .artstation import download as _
from .danbooru import download as _
from .deviantart import download as _
from .imgur import download as _
from .pixiv import download as _
from .reddit import download as _
from .twitter import download as _
from .wallhaven import download as _

def download(slug: str) -> Callable[[list[str], str], Coroutine[Any, Any, None]]: ...
def register(slug: str) -> Callable[[], None]: ...



import os.path
from collections import Counter, defaultdict, namedtuple
from enum import Enum
from functools import reduce
from urllib.parse import urlparse

from art_dl.cache import cache
from art_dl.log import Logger, Progress
from art_dl.utils.download import download_binary
from art_dl.utils.path import mkdir
from art_dl.utils.print import counter2str
from art_dl.utils.proxy import ClientSession, ProxyClientSession

SLUG = 'artstation'
BASE_URL = 'https://www.artstation.com'
USER_PROJECTS_URL = '/users/{user}/projects.json'
PROJECT_INFO_URL = '/projects/{hash}.json'

logger = Logger(inline=True)
progress = Progress()


class ParsedType(str, Enum):
	art = 'art'
	artist = 'artist'


Parsed = namedtuple('Parsed', ['id', 'type'])
Project = namedtuple('Project', ['title', 'hash_id', 'assets'])


class DownloadResult(str, Enum):
	download = 'download'
	no_image = 'no_image'
	skip = 'skip'


def parse_link(url: str) -> Parsed:
	parsed = urlparse(url)
	path = parsed.path.lstrip('/').split('/')

	if path[0] == 'artwork':
		# https://www.artstation.com/artwork/<hash>
		return Parsed(path[1], ParsedType.art)

	# https://www.artstation.com/<artist>
	return Parsed(path[0], ParsedType.artist)


async def list_projects(session: ClientSession, user: str):
	async with session.get(USER_PROJECTS_URL.format(user=user)) as response:
		return (await response.json())['data']


async def fetch_project(session: ClientSession, project: str):
	async with session.get(PROJECT_INFO_URL.format(hash=project)) as response:
		logger.info('add', project, progress=progress)
		result = await response.json()

	return {
		'assets': list({
			'has_image': a['has_image'],
			'id': a['id'],
			'image_url': a['image_url'],
			'title': a['title'],
		} for a in result['assets']),
		'hash_id': result['hash_id'],
		'title': result['title'],
		'user': {
			'username': result['user']['username']
		},
	}


async def fetch_asset(
	session: ClientSession,
	project_hash: str,
	asset,
	save_folder,
	project_prefix=None
) -> DownloadResult:
	if asset['has_image'] is False:
		logger.info('no image', project_hash)
		return DownloadResult.no_image

	asset_id = asset['id']
	# https://cdna.artstation.com/p/assets/images/images/path/to/file.jpg?1593595729 -> .jpg
	file_ext = os.path.splitext(urlparse(asset['image_url']).path.split('/')[-1])[1]
	sep = ' - '
	name = sep.join([
		asset['title'] or '',
		# if project_prefix is not empty, in collection only 1 image
		# project name written to file name
		project_prefix or '',
		str(asset_id) + file_ext
	]).strip(sep).replace(sep * 2, sep)
	filename = os.path.join(save_folder, name)

	if os.path.exists(filename):
		logger.verbose('skip existing', project_hash, asset_id, progress=progress)
		return DownloadResult.skip

	logger.info('download', project_hash, asset_id, progress=progress)
	await download_binary(session, asset['image_url'], filename)
	return DownloadResult.download


async def download(urls: list[str], data_folder: str):
	stats = Counter()  # type: ignore
	progress.total = len(urls)

	# { '<artist>': [Project(1), ...] }
	projects: dict[str, list[Project]] = defaultdict(list)

	logger.configure(prefix=[SLUG, 'queue'], inline=True)
	for url in urls:
		progress.i += 1

		parsed = parse_link(url)

		async with ProxyClientSession(BASE_URL) as session:
			if parsed.type == ParsedType.artist:
				projects_list = [p['hash_id'] for p in await list_projects(session, parsed.id)]
				stats.update(artist=1)
			elif parsed.type == ParsedType.art:
				projects_list = [parsed.id]
				stats.update(art=1)
			else:
				# this should never be called
				logger.verbose('error parsing')
				continue

			for project_hash in projects_list:
				cached: dict = cache.select(SLUG, project_hash, as_json=True)

				if cached is None:
					p = await fetch_project(session, project_hash)
					cache.insert(SLUG, project_hash, p, as_json=True)
				else:
					p = cached

				artist = p['user']['username']
				projects[artist].append(Project(p['title'], p['hash_id'], p['assets']))

	for artist in projects.keys():
		mkdir(os.path.join(data_folder, artist))

	logger.info(counter2str(stats), end='\n')

	# download assets
	logger.configure(prefix=[SLUG, 'download'], inline=True)
	stats = Counter()
	progress.i = 0
	progress.total = sum(reduce(lambda a, b: a + len(b.assets), p, 0) for p in projects.values())
	async with ProxyClientSession() as session:
		for artist, projects_list in projects.items():
			for project in projects_list:
				save_folder = os.path.join(data_folder, artist)
				sub = f"{project.title} - {project.hash_id}"

				if len(project.assets) > 1:
					# save to sub-folder
					save_folder = os.path.join(save_folder, sub)
					mkdir(save_folder)
					# do not append 'sub' to files names in sub-folder
					sub = None  # type: ignore

				for asset in project.assets:
					progress.i += 1

					res = await fetch_asset(session, project.hash_id, asset, save_folder, sub)
					stats.update({res.value: 1})

	logger.configure(prefix=[SLUG], inline=True)
	logger.info(counter2str(stats))
	logger.newline(normal=True)



import json
import os.path
from asyncio import sleep
from collections import Counter, namedtuple
from urllib.parse import parse_qs, urlparse

from aiohttp import ServerDisconnectedError
from lxml import etree

from art_dl.cache import cache
from art_dl.log import Logger, Progress
from art_dl.utils.download import download_binary
from art_dl.utils.path import filename_normalize, filename_unhide, mkdir
from art_dl.utils.print import counter2str
from art_dl.utils.proxy import ClientSession, ProxyClientSession
from art_dl.utils.url import parse_range

SLUG = 'pixiv'
HEADERS = {
	'referer': 'https://www.pixiv.net/',
}
URL = 'https://www.pixiv.net/en/artworks/'

ERROR_MESSAGES = {
	'404': 'work has been deleted or the ID does not exist'
}

logger = Logger(prefix=[SLUG, 'download'], inline=True)
progress = Progress()

Parsed = namedtuple('Parsed', ['id', 'range'], defaults=[None, None])


def parse_link(url: str):
	parsed = urlparse(url)
	path = parsed.path.lstrip('/').split('/')

	imgs_range = parse_range(parsed.fragment)
	if imgs_range is not None:
		# convert pixiv indexing, which starts from 1,
		# to indexing like range() function, which starts from 0
		imgs_range = list(map(lambda i: i - 1, imgs_range))

	if parsed.netloc == 'zettai.moe':
		# https://zettai.moe/detail?id=<id>
		query = parse_qs(parsed.query)
		return Parsed(query['id'][0], imgs_range)

	if path[1] == 'artworks':
		# https://www.pixiv.net/<lang>/artworks/<id>
		return Parsed(path[2], imgs_range)

	if path[0] == 'artworks':
		# https://www.pixiv.net/artworks/<id>
		return Parsed(path[1], imgs_range)

	return Parsed()


async def fetch_info(session: ClientSession, parsed: Parsed):
	url = URL + parsed.id
	logger.info('fetch info', parsed.id, progress=progress)
	async with session.get(url) as response:
		if response.status == 404:
			return {
				'error': 404
			}
		data = await response.text()

	root = etree.HTML(data)
	json_data = json.loads(root.xpath('//meta[@name=\'preload-data\']/@content')[0])
	art = json_data['illust'][parsed.id]

	return {
		'count': art['pageCount'],
		'first_url': art['urls']['original'],
		'id': parsed.id,
		# username can begin with a dot
		'artist': filename_unhide(filename_normalize(art['userName'])),
		'title': filename_normalize(art['title']),
	}


async def download_art(
	session: ClientSession, art_info: Parsed, info: dict, save_folder: str
) -> Counter:
	stats = Counter()  # type: ignore

	# https://i.pximg.net/img-original/img/.../xxx_p0.png
	base_url, ext = os.path.splitext(info['first_url'])
	base_url = base_url[:-1]

	total_imgs_count = info['count']
	ind_range = art_info.range or range(total_imgs_count)

	name_prefix = art_info.id + ' - ' + info['title']
	for i in ind_range:
		if i >= total_imgs_count:
			# prevent range bigger than images count
			# equal because all images indexes are in [0, 'count' - 1]
			break

		log_info = [art_info.id]
		if total_imgs_count > 1:
			# log image number only if more than one image
			log_info.extend(['/', i + 1])

		name = name_prefix + f'_p{i}' + ext
		filename = os.path.join(save_folder, name)
		if os.path.exists(filename):
			logger.verbose('skip existing', *log_info, progress=progress)
			stats.update(skip=1)
			continue

		logger.info('download', *log_info, progress=progress)
		url = base_url + str(i) + ext
		await download_binary(session, url, filename)
		stats.update(download=1)

	return stats


async def download(urls: list[str], data_folder: str):
	stats = Counter()  # type: ignore
	progress.total = len(urls)

	async with ProxyClientSession(headers=HEADERS) as session:
		for url in urls:
			progress.i += 1

			parsed = parse_link(url)
			if parsed.id is None:
				logger.warn('unsupported link:', url)
				stats.update(skip=1)
				continue

			cached: dict = cache.select(SLUG, parsed.id, as_json=True)

			if cached is None:
				info = await fetch_info(session, parsed)

				# do not cache because it can be just wrong url, not deleted
				if 'error' in info:
					logger.warn(parsed.id, 'error:', ERROR_MESSAGES[str(info['error'])])
					continue

				cache.insert(SLUG, parsed.id, info, as_json=True)
			else:
				info = cached

			save_folder = os.path.join(data_folder, info['artist'])
			mkdir(save_folder)

			while True:
				try:
					dl_stats = await download_art(session, parsed, info, save_folder)
					stats.update(dl_stats)
					break
				except ServerDisconnectedError:
					logger.info('error, retrying in 5 seconds')
					await sleep(5)

	logger.info(counter2str(stats))
	logger.newline(normal=True)



from urllib.parse import urlencode

from art_dl.redirect_server import run as run_redirect_catch_server
from art_dl.utils.credentials import creds

from .common import AUTH_LOG_PREFIX, BASE_URL, CREDS_PATHS, REDIRECT_URI, logger

AUTH_URL = BASE_URL + '/oauth2/authorize'


def ask_app_creds():
	client_id = creds.get(CREDS_PATHS.client_id)
	client_secret = creds.get(CREDS_PATHS.client_secret)

	if client_id is not None and client_secret is not None:
		ans = input('Application data already saved, again? [y/N] ')
		if ans.lower() in ['n', '']:
			return
		elif ans.lower() != 'y':
			print('What?')
			quit(1)

	creds.save(CREDS_PATHS.client_id, input('Enter client_id: '))
	creds.save(CREDS_PATHS.client_secret, input('Enter client_secret: '))


def register():
	""" Authorize application """
	ask_app_creds()

	# callback
	def cred_saver(new_code):
		creds.save(CREDS_PATHS.code, new_code)

	query = {
		'response_type': 'code',
		'client_id': creds.get(CREDS_PATHS.client_id),
		'redirect_uri': REDIRECT_URI,
		'scope': ' '.join(['browse']),
		'view': 'login'
	}
	url = f'{AUTH_URL}?{urlencode(query)}'

	try:
		run_redirect_catch_server(url, cred_saver)
	except SystemExit:
		logger.info('server stopped', prefix=AUTH_LOG_PREFIX)

	logger.info('authorized', prefix=AUTH_LOG_PREFIX)



import os.path
from collections import Counter, namedtuple
from enum import Enum
from typing import Any
from urllib.parse import urlparse

from art_dl.cache import cache
from art_dl.log import Logger, Progress
from art_dl.utils.download import download_binary
from art_dl.utils.path import filename_normalize, mkdir
from art_dl.utils.print import counter2str
from art_dl.utils.proxy import ClientSession, ProxyClientSession
from art_dl.utils.retry import retry

SLUG = 'reddit'

JSON_URI = 'https://www.reddit.com/comments/{id}.json'
IMAGE_URI = 'https://i.redd.it/'

DATA_CACHE_POSTFIX = ':data'
SKIP_CACHE_TAG = 'SKIP'

REDDIT_DOMAINS = ('reddit.com', 'i.redd.it', 'v.redd.it')

logger = Logger(prefix=[SLUG, 'download'], inline=True)
progress = Progress()

Parsed = namedtuple('Parsed', ['id'])


class DownloadResult(str, Enum):
	download = 'download'
	skip = 'skip'


def parse_link(url: str) -> Parsed:
	parsed = urlparse(url)
	path = parsed.path.lstrip('/').split('/')

	if len(path) == 1 and parsed.netloc == 'redd.it':
		# https://redd.it/<id>
		return Parsed(id=path[0])

	if len(path) == 2 and path[0] in ('comments', 'gallery'):
		# https://www.reddit.com/comments/<id>
		# https://www.reddit.com/gallery/<id>
		return Parsed(id=path[1])

	if len(path) >= 4 and path[0] == 'r' and path[2] == 'comments':
		# https://www.reddit.com/r/<subreddit>/comments/<id>/<any name>
		return Parsed(id=path[3])

	return Parsed(id=None)


async def fetch_data(session: ClientSession, url: str) -> Any:
	async with session.get(url) as response:
		response.raise_for_status()
		data = (await response.json())[0]['data']['children'][0]['data']

	media_metadata = data.get('media_metadata')
	data = {
		'domain': data['domain'],
		'is_gallery': data.get('is_gallery', False),
		'is_video': data['is_video'],
		'subreddit': data['subreddit'],
		'title': data['title'],
		'url': data['url'],
	}
	if data['is_gallery'] is True:
		data['media_ext'] = {
			# info['m'] is mime type
			media_id: info['m'].split('/')[1]
			for media_id, info in media_metadata.items()
		}
	return data


async def download_art(
	session: ClientSession,
	url: str,
	folder: str,
	name: str,
	log_name: str,
) -> DownloadResult:
	filename = os.path.join(folder, name)
	if os.path.exists(filename):
		logger.info('skip existing', log_name, progress=progress)
		return DownloadResult.skip

	logger.info('download', log_name, progress=progress)
	await download_binary(session, url, filename)
	return DownloadResult.download


async def download(urls: list[str], data_folder: str):
	stats = Counter()  # type: ignore
	progress.total = len(urls)

	sep = ' - '

	async with ProxyClientSession() as session:
		for url in urls:
			progress.i += 1

			parsed = parse_link(url)

			if parsed.id is None:
				logger.warn('unsupported link', url, progress=progress)
				stats.update(skip=1)
				continue

			cached = cache.select(SLUG, parsed.id)

			if cached == SKIP_CACHE_TAG:
				logger.verbose('skip', url, progress=progress)
				stats.update(skip_video=1)
				continue

			if cached is None:
				data = await fetch_data(session, JSON_URI.format(id=parsed.id))
				cache.insert(SLUG, parsed.id + DATA_CACHE_POSTFIX, data, as_json=True)
			else:
				data = cache.select(SLUG, parsed.id + DATA_CACHE_POSTFIX, as_json=True)

			domain = data['domain']
			if domain not in REDDIT_DOMAINS:
				logger.warn('media is from', domain, url + ':', data['url'], progress=progress)
				if domain == 'imgur.com':
					retry.add(data['url'])
					stats.update(will_retry=1)
				elif domain == 'i.imgur.com':
					imgur_id, _ = os.path.splitext(data['url'].split('/')[-1])
					retry.add('https://imgur.com/' + imgur_id)
					stats.update(will_retry=1)
				else:
					stats.update(skip=1)
				continue

			save_folder = os.path.join(data_folder, data['subreddit'])
			title = sep.join((data['title'], parsed.id))
			title = filename_normalize(title)
			is_gallery: bool = data.get('is_gallery', False)

			if is_gallery:
				folder = os.path.join(save_folder, title)
				mkdir(folder)

				i = 0
				for media_id, ext in data['media_ext'].items():
					url_filename = media_id + '.' + ext
					url = IMAGE_URI + url_filename
					res = await download_art(
						session, url, folder, url_filename, f'{parsed.id}/{media_id} - {i}'
					)
					stats.update({res.value: 1})
					i += 1

				if cached is None:
					cache.insert(SLUG, parsed.id, 'gallery')
			elif data['is_video'] is True:
				logger.verbose('skip video', url, progress=progress)
				cache.insert(SLUG, parsed.id, SKIP_CACHE_TAG)
				cache.delete(SLUG, parsed.id + DATA_CACHE_POSTFIX)
				stats.update(skip_video=1)
			else:
				url = data['url']
				url_filename = urlparse(url).path.lstrip('/')
				if cached is None:
					cache.insert(SLUG, parsed.id, 'image')

				media_id, ext = os.path.splitext(url_filename)
				filename = sep.join((title, media_id)) + ext
				mkdir(save_folder)
				res = await download_art(
					session, url, save_folder, filename, f'{parsed.id}/{media_id}'
				)
				stats.update({res.value: 1})

	logger.info(counter2str(stats))
	logger.newline(normal=True)



import os.path
from collections import Counter, namedtuple
from enum import Enum
from typing import Any
from urllib.parse import urlparse

from art_dl.cache import cache
from art_dl.log import Logger, Progress
from art_dl.utils.download import download_binary
from art_dl.utils.path import filename_normalize, mkdir
from art_dl.utils.print import counter2str
from art_dl.utils.proxy import ClientSession, ProxyClientSession

SLUG = 'imgur'
API_URL = 'https://api.imgur.com/3/{type}/{id}'
# client_id just from devtools
HEADERS = {
	'authorization': 'Client-ID 546c25a59c58ad7'
}

logger = Logger(prefix=[SLUG, 'download'], inline=True)
progress = Progress()

Parsed = namedtuple('Parsed', ['id', 'type'], defaults=[None, None])


class LinkType(str, Enum):
	album = 'album'
	image = 'image'


class DownloadResult(str, Enum):
	download = 'download'
	skip = 'skip'


def parse_link(url: str):
	parsed = urlparse(url)
	path = parsed.path.lstrip('/').split('/')

	if path[0] == 'a' or path[0] == 'gallery':
		# https://imgur.com/a/<id>
		# https://imgur.com/gallery/<id>
		return Parsed(path[1], LinkType.album)

	if path[0] == 't':
		# https://imgur.com/t/<tag>/<id>
		return Parsed(path[2], LinkType.album)

	if len(path) == 1:
		# https://imgur.com/<id>
		return Parsed(path[0], LinkType.image)

	return Parsed()


async def fetch_info(session: ClientSession, album: Parsed) -> Any:
	logger.verbose('fetch info', album.id, progress=progress)

	url = API_URL.format(id=album.id, type=album.type)
	async with session.get(url, headers=HEADERS) as response:
		response.raise_for_status()
		info = (await response.json())['data']

	if info.get('is_album', False) is False:
		# the same model, so just put to array
		info['images'] = [info]

	return {
		'id': info['id'],
		'title': info['title'] or '',
		'images': list({
			'id': image['id'],
			'link': image['link'],
			'ext': os.path.splitext(image['link'])[-1],
			'title': image['title'] or '',
		} for image in info['images']),
	}


async def download_art(
	session: ClientSession, link: str, save_folder: str, name: str
) -> DownloadResult:
	filename = os.path.join(save_folder, name)
	if os.path.exists(filename):
		logger.verbose('skip existing', name, progress=progress)
		return DownloadResult.skip

	logger.info('download', name, progress=progress)
	await download_binary(session, link, filename)
	return DownloadResult.download


async def download(urls: list[str], data_folder: str):
	stats = Counter()  # type: ignore
	progress.total = len(urls)

	sep = ' - '

	async with ProxyClientSession() as session:
		for url in urls:
			progress.i += 1

			parsed = parse_link(url)

			if parsed.id is None:
				logger.warn('unsupported link', url)
				stats.update(skip=1)
				continue

			cached = cache.select(SLUG, parsed.id, as_json=True)

			if cached is None:
				info = await fetch_info(session, parsed)
				cache.insert(SLUG, parsed.id, info, as_json=True)
			else:
				info = cached

			images = info['images']
			one_image = len(images) == 1
			title_prefix = sep.join((info['title'], info['id'])).strip(sep)
			title_prefix = filename_normalize(title_prefix)

			if one_image:
				save_folder = data_folder
			else:
				# save to sub-folder
				save_folder = os.path.join(data_folder, title_prefix)
				title_prefix = ''
			mkdir(save_folder)

			for image in images:
				title = (
					sep.join((title_prefix, image['title'], image['id'])
								).strip(sep).replace(sep * 2, sep)
				)
				name = title + image['ext']
				res = await download_art(session, image['link'], save_folder, name)
				stats.update({res.value: 1})

	logger.info(counter2str(stats))
	logger.newline(normal=True)



import os.path
from asyncio import sleep
from collections import Counter, namedtuple
from enum import Enum
from glob import glob
from typing import Any, Tuple
from urllib.parse import urlparse

from art_dl.cache import cache
from art_dl.log import Logger, Progress
from art_dl.utils.credentials import creds
from art_dl.utils.download import download_binary
from art_dl.utils.path import filename_normalize, filename_shortening, mkdir
from art_dl.utils.print import counter2str
from art_dl.utils.proxy import ClientSession, ProxyClientSession

SLUG = 'wallhaven'
CREDS_PATH = [SLUG, 'api_key']

API_URL = 'https://wallhaven.cc/api/v1/w/'

logger = Logger(prefix=[SLUG, 'download'], inline=True)
progress = Progress()


class FetchDataAction(Enum):
	download = 0
	retry_with_key = 1
	skip = 2


Parsed = namedtuple('Parsed', ['id'])


def parse_link(url: str):
	parsed = urlparse(url)
	path = parsed.path.lstrip('/').split('/')

	if parsed.netloc == 'wallhaven.cc':
		# https://wallhaven.cc/w/<id>
		# strip 'w'
		path.pop(0)

	# https://whvn.cc/<id>
	return Parsed(path[0])


async def fetch_data(
	session: ClientSession,
	img_id: str,
	params: dict,
	with_key: bool,
	has_api_key: bool,
) -> Tuple[Any, FetchDataAction]:
	# loop for retrying on rate limit
	while True:
		async with session.get(API_URL + img_id, params=params) as response:
			if response.status == 429:
				logger.info('to many requests, sleeping for 10 seconds', progress=progress)
				await sleep(10)
				continue
			elif response.status == 401:
				if with_key:
					logger.warn('invalid api_key, skip')
					return None, FetchDataAction.skip

				if has_api_key:
					logger.verbose('queueing NSFW', img_id)
					return None, FetchDataAction.retry_with_key

				logger.warn('skip NSFW', img_id, '(api_key not present)')
				return None, FetchDataAction.skip

			data = (await response.json())['data']
			data = {
				'id': data['id'],
				'path': data['path'],
				'tags': list(t['name'] for t in data['tags'])
			}
			return data, FetchDataAction.download


async def download(urls: list[str], data_folder: str, with_key=False):
	mkdir(data_folder)

	stats = Counter()  # type: ignore
	progress.set(0, len(urls))

	retry_with_key = []

	api_key = creds.get(CREDS_PATH)
	has_api_key = api_key is not None
	params = {}
	if with_key:
		if has_api_key:
			logger.info('using api_key', end='\n')
			params = {
				'apikey': api_key
			}
		else:
			logger.warn('you should add api_key')
			return

	async with ProxyClientSession() as session:
		for url in urls:
			progress.i += 1
			should_skip = False

			parsed = parse_link(url)
			existing = glob(f'{data_folder}/{parsed.id} - *.*')
			if len(existing) == 1:
				logger.verbose('skip existing', parsed.id)
				should_skip = True
			elif len(existing) > 1:
				logger.warn('duplicated files for art', parsed.id)
				should_skip = True

			if url in retry_with_key:
				logger.verbose('duplicate link', url)
				should_skip = True

			if should_skip:
				stats.update(skip=1)
				continue

			cached = cache.select(SLUG, parsed.id, as_json=True)

			if cached is None:
				data, action = await fetch_data(session, parsed.id, params, with_key, has_api_key)

				if action == FetchDataAction.retry_with_key:
					stats.update(will_retry=1)
					retry_with_key.append(url)
				elif action == FetchDataAction.skip:
					stats.update(skip=1)
				if action != FetchDataAction.download:
					continue

				cache.insert(SLUG, parsed.id, data, as_json=True)
			else:
				data = cached

			logger.info('download', data['id'], progress=progress)

			full_url = data['path']
			name = data['id'] + ' - ' + ', '.join(data['tags'])
			name = filename_normalize(name) + os.path.splitext(full_url)[1]
			name = filename_shortening(name, with_ext=True)
			filename = os.path.join(data_folder, name)

			await download_binary(session, full_url, filename)
			stats.update(download=1)

	logger.info(counter2str(stats))
	logger.newline(normal=True)

	if len(retry_with_key) > 0:
		return await download(retry_with_key, data_folder, True)


def register():
	"""Ask key"""
	api_key = creds.get(CREDS_PATH)
	if api_key is not None:
		ans = input('Key already saved, again? [y/N] ')
		if ans.lower() in ['n', '']:
			return
		elif ans.lower() != 'y':
			print('What?')
			quit(1)
	creds.save(CREDS_PATH, input('Enter api_key: '))
	logger.info('saved')



from art_dl.utils.db import DB
from art_dl.utils.dirs import DIRS

CONFIG_DB = DIRS.config + '/config.db'
CONFIG_TABLE = 'configuration'


class Config():

	def __init__(self) -> None:
		self.db = DB(CONFIG_DB, CONFIG_TABLE)

	def get(self, key: str, default=None):
		return self.db.select(key) or default

	def set(self, key: str, value: str):
		self.db.insert(key, value)

	def check_value(self, key: str):
		value = self.get(key)
		if value is not None:
			print('Current value:', value)

	def input_entry(self, name=None):
		key = name or input('Enter key: ')
		if key is not None:
			self.check_value(key)

		value = input(f'Enter {name or "value"}: ')

		self.set(key, value)


config = Config()



'''https://danbooru.donmai.us/wiki_pages/help:api'''
from art_dl.utils.proxy import ClientSession, ProxyClientSession


async def fetch_smth(session: ClientSession, url: str):
	async with session.get(url) as response:
		print(response)


async def download(urls: list[str], data_folder: str):
	async with ProxyClientSession() as session:
		for url in urls:
			await fetch_smth(session, url)



import os.path
from collections import Counter, namedtuple
from enum import Enum
from urllib.parse import unquote, urljoin, urlparse

from aiohttp import ClientTimeout
from aiohttp.client_exceptions import ServerTimeoutError
from lxml import etree

from art_dl.cache import cache
from art_dl.log import Logger, Progress
from art_dl.utils.download import download_binary
from art_dl.utils.path import filename_normalize, filename_shortening, mkdir
from art_dl.utils.print import counter2str
from art_dl.utils.proxy import ClientSession, ProxyClientSession

SLUG = 'twitter'

BASE_URL = 'https://nitter.net'
# https://github.com/libredirect/libredirect/blob/master/src/instances/data.json
# except https://github.com/libredirect/libredirect/blob/master/src/instances/blacklist.json
FALLBACK_URLS = (
	'https://nitter.42l.fr',
	'https://nitter.pussthecat.org',
	'https://nitter.fdn.fr',
	'https://nitter.1d4.us',
	'https://nitter.kavin.rocks',
	'https://nitter.unixfox.eu',
	# return to default
	'https://nitter.net',
)
# ind in FALLBACK_URLS
CURRENT_URL_IND = -1

COOKIES = {
	# hide replies and images in replies
	'hideReplies': 'on',
	# just reduce response a little
	'hideTweetStats': 'on',
}
HEADERS = {
	# why Title case??
	'User-Agent': 'Mozilla/5.0'
}
SESSION_TIMEOUT = ClientTimeout(sock_read=10)

Parsed = namedtuple('Parsed', ['id', 'account', 'path'], defaults=[None, None, None])

logger = Logger(prefix=[SLUG, 'download'], inline=True)
progress = Progress()


class DownloadResult(str, Enum):
	download = 'download'
	skip = 'skip'


def switch_instance():
	global BASE_URL
	global CURRENT_URL_IND

	CURRENT_URL_IND = (CURRENT_URL_IND + 1) % len(FALLBACK_URLS)
	BASE_URL = FALLBACK_URLS[CURRENT_URL_IND]

	logger.warn('connect timeout, switching instance to', BASE_URL)


def parse_link(url: str) -> Parsed:
	parsed = urlparse(url)
	original_path = parsed.path
	path = original_path.lstrip('/').split('/')

	if len(path) > 2 and path[1] == 'status':
		# https://(mobile.)twitter.com/<account>/status/<id>
		return Parsed(path[2], path[0], original_path)

	return Parsed()


async def fetch_info(session: ClientSession, parsed: Parsed):
	logger.info('fetch info', f'{parsed.account}/{parsed.id}', progress=progress)
	try:
		# wait for api: https://github.com/zedeus/nitter/issues/192
		async with session.get(urljoin(BASE_URL, parsed.path)) as response:
			data = await response.text()
	except ServerTimeoutError:
		switch_instance()
		return await fetch_info(session, parsed)

	root = etree.HTML(data)
	description = root.xpath('//meta[@property=\'og:description\']/@content')[0]
	images_urls = root.xpath(
		'//div[@class="attachments"]/div/div[@class="attachment image"]/a/@href'
	)

	return {
		'description': description,
		# save original url (non unquoted) bc when make request like /pic/media/...?name=orig
		# it will return not original image, but if make request /pic/media%2F...%3Fname%3Dorig
		# it will return original image
		'images': [{
			'url': i,
			'ext': os.path.splitext(urlparse(unquote(i)).path)[1],
		} for i in images_urls],
		'count': len(images_urls),
	}


async def download_image(
	session: ClientSession, url: str, save_folder: str, name: str, log_info: str
) -> DownloadResult:
	filename = os.path.join(save_folder, name)
	if os.path.exists(filename):
		logger.verbose('skip existing', log_info, progress=progress)
		return DownloadResult.skip

	logger.info('download', log_info, progress=progress)
	try:
		await download_binary(session, urljoin(BASE_URL, url), filename)
	except ServerTimeoutError:
		switch_instance()
		return await download_image(session, url, save_folder, name, log_info)

	return DownloadResult.download


async def download(urls: list[str], data_folder: str):
	stats = Counter()  # type: ignore
	progress.total = len(urls)
	sep = ' - '

	async with ProxyClientSession(cookies=COOKIES, timeout=SESSION_TIMEOUT, headers=HEADERS) as session:
		for url in urls:
			progress.i += 1

			parsed = parse_link(url)
			if parsed.id is None:
				logger.warn('unsupported link:', url)
				stats.update(skip=1)
				continue

			cache_key = parsed.account + ':' + parsed.id
			cached: dict = cache.select(SLUG, cache_key, as_json=True)

			if cached is None:
				info = await fetch_info(session, parsed)
				cache.insert(SLUG, cache_key, info, as_json=True)
			else:
				info = cached

			title_prefix = sep.join((parsed.account, parsed.id, info['description'])).strip(sep)
			# 245 = 255 - len('.xxxx') - len(' - xx')
			title_prefix = filename_shortening(filename_normalize(title_prefix), 245)
			add_index = (info['count']) > 1

			save_folder = os.path.join(data_folder, parsed.account)
			mkdir(save_folder)

			i = 0
			for image in info['images']:
				filename = (title_prefix + sep + str(i)) if add_index else title_prefix
				filename += image['ext']
				i += 1

				log_info = f'{parsed.account}/{parsed.id}'
				if add_index:
					log_info += sep + str(i)
				res = await download_image(session, image['url'], save_folder, filename, log_info)
				stats.update({res.value: 1})

	logger.info(counter2str(stats))
	logger.newline(normal=True)



from art_dl.cache import cache


class Retry:
	KEY = 'RETRY'

	_old_list: list[str] = []

	def get(self) -> list[str] | None:
		return cache.select(None, self.KEY, as_json=True)

	def add(self, to_retry: list[str] | str):
		urls = self.get() or []
		if isinstance(to_retry, list):
			urls.extend(to_retry)
		else:
			urls.append(to_retry)

		cache.insert(None, self.KEY, urls, as_json=True)

	def clear(self, *, force=False):
		if force:
			self._old_list = []
		else:
			# save to prevent urls lost, for example on crash
			# replace old list instead of extend to not saving old urls
			self._old_list = self.get() or []

		cache.delete(None, self.KEY)

	def __del__(self):
		if len(self._old_list) > 0:
			self.add(self._old_list)


retry = Retry()



from art_dl.utils.db import DB
from art_dl.utils.dirs import DIRS

CREDS_DB = DIRS.config + '/config.db'
CREDS_TABLE = 'credentials'


class Credentials:

	def __init__(self) -> None:
		self.db = DB(CREDS_DB, CREDS_TABLE)

	@staticmethod
	def _key(path: list[str]):
		return '.'.join(path)

	def get(self, path: list[str]):
		return self.db.select(self._key(path))

	def save(self, path: list[str], value: str):
		self.db.insert(self._key(path), value)

	def delete(self, path: list[str]):
		self.db.delete(self._key(path))


creds = Credentials()



import platformdirs

from art_dl.utils.path import mkdir

appname = 'artworks-downloader'
appauthor = 'istudyatuni'


class DIRS:
	cache = platformdirs.user_cache_dir(appname, appauthor)
	config = platformdirs.user_config_dir(appname, appauthor)


mkdir(DIRS.cache)
mkdir(DIRS.config)



import sqlite3 as sql
from json import dumps, loads
from typing import Any


class Queries:
	init = '''CREATE TABLE IF NOT EXISTS {table} (
		key TEXT NOT NULL PRIMARY KEY,
		value TEXT
	)'''
	insert = '''INSERT OR IGNORE INTO {table} (key, value) VALUES (:key, :value)'''
	select = '''SELECT value FROM {table} WHERE key = :key'''
	delete = '''DELETE FROM {table} WHERE key = :key'''

	def __init__(self, table: str) -> None:
		for q in ['init', 'insert', 'select', 'delete']:
			self.__setattr__(q, self.__getattribute__(q).format(table=table))


class DB:
	""" Key-value sqlite wrapper """

	def __init__(self, db_name: str, table: str) -> None:
		self.db_name = db_name
		self.queries = Queries(table)
		self.connect()

	def connect(self):
		self.conn = sql.connect(self.db_name)
		self.conn.row_factory = sql.Row
		self.cursor = self.conn.cursor()

		self.cursor.executescript(self.queries.init)
		self.conn.commit()

	def insert(self, key: str, value: str | Any, *, as_json=False):
		# if not as json value should be a string
		if as_json is False and not isinstance(value, str):
			raise TypeError('Value should be a string')

		self.cursor.execute(
			self.queries.insert, {
				'key': key,
				'value': dumps(value) if as_json else value,
			}
		)
		self.conn.commit()

	def select(self, key: str, *, as_json=False):
		res = self.cursor.execute(self.queries.select, {
			'key': key
		}).fetchone()
		if res is None:
			return res
		value = res['value']
		return loads(value) if as_json else value

	def delete(self, key: str):
		self.cursor.execute(self.queries.delete, { 'key': key })
		self.conn.commit()



from aiofiles import open as aopen
from aiohttp import ClientSession

from art_dl.utils.cleanup import cleanup


async def download_binary(session: ClientSession, url: str, filename: str):
	async with session.get(url, raise_for_status=True) as response:
		cleanup.set(filename)
		async with aopen(filename, 'wb') as file:
			try:
				await file.write(await response.read())
				cleanup.forget()
			except:
				cleanup.clean()
				print('REMOVING EMPTY FILE')
				raise



from aiohttp import ClientSession
from aiohttp_socks import ProxyConnector  # type: ignore

from art_dl.utils.config import config

__all__ = ['ClientSession', 'ProxyClientSession']


def _can_use_proxy_url(url: str | None):
	return url is not None and url != ''


class ProxyClientSession(ClientSession):

	def __init__(self, *args, **kwargs):
		proxy_url = config.get('proxy')
		if kwargs.get('connector') is None and _can_use_proxy_url(proxy_url):
			kwargs['connector'] = ProxyConnector.from_url(proxy_url)
			# print('proxy', proxy_url)

		super().__init__(*args, **kwargs)

	# async def _request(self, *args, **kwargs):
	# 	print('_request', *args)
	# 	return await super()._request(*args, **kwargs)



import os

from art_dl.cache import CACHE_DB
from art_dl.utils.db import DB


class Cleanup:
	KEY = 'filename'

	def __init__(self) -> None:
		self.db = DB(CACHE_DB, 'cleanup')

	def set(self, filename: str):
		""" Remember file for cleaning """
		self.db.insert(self.KEY, filename)

	def forget(self):
		self.db.delete(self.KEY)

	def clean(self):
		""" Perform cleanup """
		filename = self.db.select(self.KEY)
		if filename is None:
			return
		if os.path.exists(filename):
			os.remove(filename)
		self.forget()


cleanup = Cleanup()



from typing import Any

from art_dl.utils.db import DB
from art_dl.utils.dirs import DIRS

CACHE_DB = DIRS.cache + '/cache.db'


class Cache:

	def __init__(self) -> None:
		self.db = DB(CACHE_DB, 'cache')

	@staticmethod
	def _key(slug: str | None, key: str):
		return key if slug is None else slug + ':' + key

	def insert(self, slug: str | None, key: str, value: str | Any, *, as_json=False):
		self.db.insert(self._key(slug, key), value, as_json=as_json)

	def select(self, slug: str | None, key: str, *, as_json=False):
		return self.db.select(self._key(slug, key), as_json=as_json)

	def delete(self, slug: str | None, key: str):
		self.db.delete(self._key(slug, key))


cache = Cache()



from art_dl.log import Logger, Progress

SLUG = 'deviantart'
BASE_URL = 'https://www.deviantart.com'
REDIRECT_URI = 'http://localhost:23445'


class CREDS_PATHS:
	OAUTH_KEY = 'oauth2'

	client_id = [SLUG, 'client_id']
	client_secret = [SLUG, 'client_secret']
	code = [SLUG, OAUTH_KEY, 'code']
	access_token = [SLUG, OAUTH_KEY, 'access_token']
	refresh_token = [SLUG, OAUTH_KEY, 'refresh_token']


AUTH_LOG_PREFIX = [SLUG, 'auth']

logger = Logger(prefix=[SLUG, 'download'], inline=True)
progress = Progress()


def make_cache_key(username: str, url: str):
	return ':'.join([
		'art',
		username.lower(),
		'deviationid',
		url.split('/')[-1],
	])



"""
Module to catch redirect oauth callback
"""

from typing import Any, Callable

from aiohttp import web
from aiohttp.typedefs import Handler

saver_func = lambda _: None


@web.middleware
async def middleware(request: web.Request, handler: Handler):
	resp = await handler(request)

	await resp.prepare(request)
	await resp.write_eof()

	# now hardcoded for deviantart
	if resp.status == 200:
		saver_func(request.query['code'])
		# stop server
		raise SystemExit

	return resp


async def redirect_handler(request: web.Request):
	if 'error' in request.query:
		return web.Response(
			text='An error occurred: ' + request.query['error_description'], status=400
		)
	else:
		return web.Response(text='Authorized', status=200)


def run(url: str, saver: Callable[[Any], None]):
	global saver_func
	saver_func = saver

	print('Authorizing\nOpen', url)

	app = web.Application(middlewares=[middleware])
	app.router.add_get('/', redirect_handler)
	web.run_app(app, host='localhost', port=23445)


if __name__ == '__main__':
	print('Do not run this directly')
	quit(1)



import os.path
from argparse import ArgumentParser
from asyncio import create_task, new_event_loop, set_event_loop
from typing import Optional, Tuple
from urllib.parse import urlparse

from art_dl.log import Logger, set_verbosity
from art_dl.sites import download, register
from art_dl.utils.cleanup import cleanup
from art_dl.utils.config import config
from art_dl.utils.retry import retry

SLUGS_MAPPING = {
	'artstation': ['www.artstation.com'],
	'danbooru': ['danbooru.donmai.us', 'safebooru.donmai.us'],
	'deviantart': ['www.deviantart.com'],
	'imgur': ['imgur.com'],
	'pixiv': ['www.pixiv.net', 'zettai.moe'],
	'reddit': ['redd.it', 'www.reddit.com'],
	'twitter': ['mobile.twitter.com', 'nitter.net', 'twitter.com'],
	'wallhaven': ['wallhaven.cc', 'whvn.cc'],
}

SLUGS = { url: slug
			for slug, urls in SLUGS_MAPPING.items() for url in urls }

logger = Logger(prefix=['main'])


def detect_site(url: str) -> str | None:
	return SLUGS.get(urlparse(url).netloc)


def parse_args():
	parser = ArgumentParser(description='Artworks downloader')

	parser.add_argument('-u', '--url', type=str, help='URL to download')
	parser.add_argument(
		'-l',
		'--list',
		type=str,
		help='File with list of URLs to download, one URL per line',
		default=None
	)
	parser.add_argument(
		'--folder', type=str, help='Folder to save artworks. Default folder - data', default='data'
	)

	parser.add_argument('--action', type=str, default=None)

	parser.add_argument('-q', '--quiet', action='store_true', help='Do not show logs')
	parser.add_argument('-v', '--verbose', action='store_true', help='Show more logs')

	parser.add_argument('--version', action='store_true', help='Show version')

	return parser.parse_args()


async def process_list(urls: list[str], folder: str):
	if len(urls) == 0:
		logger.info('list is empty')
		return
	if len(urls) == 1 and urls[0] is None:
		logger.info('no link')
		return

	mapping: dict[str, list[str]] = { s: []
										for s in SLUGS.values() }
	for u in urls:
		site_slug = detect_site(u)
		if site_slug is None:
			logger.info('unknown link', u)
			continue

		mapping[site_slug].append(u)

	logger.info('saving to', folder)
	tasks = []
	for slug, l in mapping.items():
		if len(l) == 0:
			continue

		save_folder = os.path.join(folder, slug)
		tasks.append(create_task(download(slug)(l, save_folder)))

	for task in tasks:
		await task


def prepare() -> Optional[Tuple[list[str], str]]:
	args = parse_args()
	# put to list for handling single url as list when download
	to_dl = [args.url]
	urls_file = args.list
	folder = os.path.abspath(args.folder)
	action = tuple(args.action.split(':')) if args.action else None

	# version
	if args.version:
		from importlib import metadata
		try:
			print(metadata.version('art-dl'))
		except metadata.PackageNotFoundError:
			if 'site-packages' not in __file__:
				logger.info('running in dev mode')
			else:
				# I don't know how is it possible but anyway
				raise
		return None

	# log config
	if args.quiet and args.verbose:
		print('You must specify either --verbose or --quiet, not both')
		quit(1)
	set_verbosity(args.quiet, args.verbose)

	# actions
	if action == ('deviantart', 'register'):
		register('deviantart')()
		return None
	elif action == ('wallhaven', 'key'):
		register('wallhaven')()
		return None
	elif action == ('config', 'proxy'):
		config.input_entry('proxy')
		return None
	elif action is not None:
		logger.info('unknown action:', args.action)
		return None

	if urls_file is not None:
		with open(urls_file) as file:
			content = file.read().strip().split('\n')
			to_dl = list(map(lambda s: s.strip(), filter(lambda e: bool(e), content)))

	return to_dl, folder


def run(urls: list[str], folder: str):
	loop = new_event_loop()
	set_event_loop(loop)
	loop.run_until_complete(process_list(urls, folder))


def _real_main():
	if (result := prepare()) is None:
		quit(0)

	urls, folder = result

	run(urls, folder)
	while (to_retry := retry.get()) is not None:
		logger.info('retrying', len(to_retry), 'urls')
		retry.clear()
		run(to_retry, folder)

	retry.clear(force=True)


def main():
	cleanup.clean()
	try:
		_real_main()
	except KeyboardInterrupt:
		logger.configure(inline=True)
		logger.warn('interrupted by user, exiting')
	cleanup.clean()


if __name__ == '__main__':
	main()



import os.path
import re
from functools import partial
from os import makedirs

mkdir = partial(makedirs, exist_ok=True)


def filename_unhide(filename: str):
	return '_' + filename if filename.startswith('.') else filename


def filename_normalize(filename: str):
	""" Normalize filename: replace `<` `>` `:` `"` `\\` `/` `|` `?` `*` with `_` """
	return re.sub(r'[<>:"\\\/|?*]', '_', filename)


def _filename_shortening_ascii(filename: str, length=255, with_ext=False):
	if with_ext:
		file, ext = os.path.splitext(filename)
		return file[:length - len(ext)] + ext

	return filename[:length]


def _filename_shortening_unicode(filename: str, length=255, with_ext=False):
	""" Strip filename to `length` bytes """
	new_filename = ''
	ext = ''
	bytes_len = 0

	if with_ext:
		filename, ext = os.path.splitext(filename)
		length -= len(ext)

	for ch in filename:
		bytes_len += len(bytes(ch, encoding='utf-8'))
		if bytes_len > length:
			break

		new_filename += ch

	return new_filename + ext


def filename_shortening(filename: str, length=255, *, with_ext=False):
	"""
	Strip filename to `length` symbols if it's a valid ascii string
	and to `length` bytes if a string contains unicode characters
	"""
	if len(bytes(filename, encoding='utf-8')) == len(filename):
		return _filename_shortening_ascii(filename, length, with_ext)

	return _filename_shortening_unicode(filename, length, with_ext)



from shutil import get_terminal_size
from typing import Optional


def print_inline_end(*values: object, sep=None, end=None):
	print(*values, sep=sep, end=end or '', flush=True)


_verbose = False
_quiet = False


def set_verbosity(q: bool = False, v: bool = False):
	global _quiet
	global _verbose
	_quiet = q
	_verbose = v
	if _verbose and _quiet:
		# is it ok to do that?
		raise Exception('log configuration error: both quiet and verbose are True')


class Progress:
	i: int = 0
	total: int = 0

	def set(self, i: int, total: int):
		self.i = i
		self.total = total

	def __str__(self) -> str:
		return f'{self.i}/{self.total}'


class Logger:
	_log_prefix_str = None
	_old_prefix_str = None
	_inline = False

	def __init__(self, *, prefix=None, inline=False) -> None:
		self._inline = inline
		if prefix is not None:
			self.configure(prefix=prefix)

	def _save_prefix(self, new_prefix: list[str]):
		self._old_prefix_str = self._log_prefix_str
		self.configure(prefix=new_prefix)

	def _restore_prefix(self):
		self._log_prefix_str = self._old_prefix_str
		self._old_prefix_str = None

	def configure(self, *, prefix: list[str] | None = None, inline: bool | None = None):
		if inline is not None:
			self._inline = inline
		if prefix and len(prefix) > 0:
			self._log_prefix_str = '[' + ']['.join(prefix) + ']'

	@property
	def _print_func(self):
		return print_inline_end if self._inline else print

	@property
	def _log_prefix(self):
		if self._log_prefix_str is not None:
			prefix = '\r' if self._inline else ''
			return prefix + self._log_prefix_str

	@staticmethod
	def _term_width():
		return get_terminal_size().columns

	@staticmethod
	def _make_print_str(*values: object, sep=None) -> str:
		# this function is for reduce number of the same calculations
		return (sep or ' ').join(str(v) for v in values)

	def _print(
		self,
		*values: object,
		progress: Optional[Progress] = None,
		sep=None,
		end=None,
		prefix: list[str] | None = None,
	):
		if _verbose:
			end = end if end else '\n'

		if prefix is not None:
			self._save_prefix(prefix)

		to_print = list(values)
		if progress is not None:
			to_print.insert(0, f'({progress})')
		if self._log_prefix is not None:
			to_print.insert(0, self._log_prefix)
		to_print_str = self._make_print_str(*to_print, sep=sep)

		spaces_offset = self._term_width() - len(to_print_str)
		if self._inline:
			# otherwise when not inline an extra line will be printed
			spaces_offset += 1

		self._print_func(to_print_str + ' ' * spaces_offset, end=end)

		if prefix is not None:
			self._restore_prefix()

	def info(
		self,
		*values: object,
		progress: Optional[Progress] = None,
		sep=None,
		end=None,
		prefix: list[str] | None = None,
	):
		if _quiet:
			return

		self._print(*values, progress=progress, sep=sep, end=end, prefix=prefix)

	def verbose(
		self,
		*values: object,
		progress: Optional[Progress] = None,
		sep=None,
		end=None,
		prefix: list[str] | None = None,
	):
		if not _verbose:
			return

		self._print(*values, progress=progress, sep=sep, end=end, prefix=prefix)

	def warn(
		self,
		*values: object,
		progress: Optional[Progress] = None,
		sep=None,
		end='\n',
		prefix: list[str] | None = None,
	):
		self._print(*values, progress=progress, sep=sep, end=end, prefix=prefix)

	@staticmethod
	def newline(*, quiet=False, verbose=False, normal=False):
		"""
		Pass `quiet=True` to print when quiet enabled, `verbose=True` to print
		when verbose enabled, `normal=True` when not quiet and not verbose enabled
		"""
		if ((quiet and _quiet) or (verbose and _verbose)
			or (normal and _quiet is False and _verbose is False)):
			print()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['art_dl', 'art_dl.sites', 'art_dl.sites.deviantart', 'art_dl.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.8.0,<0.9.0',
 'aiohttp-socks[asyncio]>=0.7.1,<0.8.0',
 'aiohttp>=3.8.1,<4.0.0',
 'lxml>=4.8.0,<5.0.0',
 'platformdirs>=2.5.2,<3.0.0']

entry_points = \
{'console_scripts': ['art-dl = art_dl:main']}

setup_kwargs = {
    'name': 'art-dl',
    'version': '0.2.7',
    'description': 'Artworks downloader',
    'long_description': '# Artworks downloader\n\n[![PyPI](https://img.shields.io/pypi/v/art-dl)](https://pypi.org/project/art-dl)\n\n## Supported sites\n\n- artstation.com [#usage](#sites-with-simple-usage)\n- deviantart.com [#usage](#deviantart)\n- imgur.com [#usage](#sites-with-simple-usage)\n- pixiv.net [#usage](#sites-with-simple-usage) [#notes](#pixiv)\n  - zettai.moe\n- reddit.com [#usage](#sites-with-simple-usage)\n- twitter.com [#usage](#sites-with-simple-usage) [#notes](#twitter)\n- wallhaven.cc [#usage](#sites-with-simple-usage) [#notes](#wallhaven)\n\n[Supported URL types](#supported-url-types)\n\n## Install\n\n### With `pip`\n\n```sh\npip install -U art-dl\n```\n\nThen run as `art-dl`: [#usage](#usage)\n\n### Build from source\n\nYou need poetry, [install](https://python-poetry.org/docs/#installation) it, then run from inside the project\n\n```sh\npoetry install --no-dev\n\n# run with poetry\npoetry run python -m art_dl\n\n# enter venv created with poetry\npoetry shell\n# run inside it as\npython -m art_dl\n# or\nart-dl\n\n# create and activate venv, e.g. with virtualenvwrapper, then\npoetry install --no-dev\nart-dl\n```\n\nAlternatively, build binary with [nuitka](https://github.com/Nuitka/Nuitka):\n\n```sh\npoetry install\npoetry shell\npython -m nuitka art_dl\n```\n\nAfter that you will have binary in the root directory.\n\n## Usage\n\n```\nusage: art-dl [-h] [-u URL] [-l LIST] [--folder FOLDER] [--action ACTION] [-q] [-v] [--version]\n\nArtworks downloader\n\noptions:\n  -h, --help            show this help message and exit\n  -u URL, --url URL     URL to download\n  -l LIST, --list LIST  File with list of URLs to download, one URL per line\n  --folder FOLDER       Folder to save artworks. Default folder - data\n  --action ACTION\n  -q, --quiet           Do not show logs\n  -v, --verbose         Show more logs\n  --version             Show version\n```\n\n### Sites with simple usage\n\nJust run\n\n```sh\n# download single url\nart-dl -u [URL]\n# download urls from file with list of urls, one url per line\nart-dl -l [file with list of urls]\n```\n\n### DeviantArt\n\nYou should have deviantart.com account, login to it, then\n\n- register an application\n  - go to https://www.deviantart.com/developers/apps\n  - click "Register Application"\n  - in field "OAuth2 Redirect URI Whitelist (Required)" under "Application Settings" block paste `http://localhost:23445`\n  - scroll to bottom and check "I have read and agree to the API License Agreement."\n  - click "Save"\n  - in the block with newly created application click "Publish"\n\n- save `client_id` and `client_secret` in this application\n  - run\n\n  ```sh\n  art-dl --action deviantart:register\n  ```\n\n  - paste needed values\n\n- authorize application\n  - open suggested link\n  - click "Authorize"\n\nAfter that you can use it as other sites: [#usage](#sites-with-simple-usage)\n\n### Proxy\n\nRun\n\n```sh\nart-dl --action config:proxy\n```\n\nEnter proxy, for example, `socks5://localhost:1080`\n\n## Notes\n\n### Pixiv\n\nIf the artwork has more one image, you can specify which images should be downloaded, for example, if the artwork has 10 images and you want to download 1, 3, 4, 5 and 7 image, you can add `#1,3-5,7` to the link for that: `https://www.pixiv.net/<lang>/artworks/<id>#1,3-5,7`.\n\n### Twitter\n\nHere we use an alternative frontend for Twitter: https://nitter.net ([Github](https://github.com/zedeus/nitter))\n\n### Wallhaven\n\nNSFW images supported only with API key, to use it, get it from [account settings](https://wallhaven.cc/settings/account), then run\n\n```sh\nart-dl --action wallhaven:key\n```\n\n## Supported URL types\n\n- **artstation.com**\n  - `https://www.artstation.com/artwork/<hash>`\n  - `https://www.artstation.com/<artist>`\n- **deviantart.com**\n  - All deviations\n    - `https://www.deviantart.com/<artist>`\n    - `https://www.deviantart.com/<artist>/gallery/all`\n  - "Featured" collection\n    - `https://www.deviantart.com/<artist>/gallery`\n  - `https://www.deviantart.com/<artist>/gallery/<some number>/<gallery name>`\n  - `https://www.deviantart.com/<artist>/art/<name>`\n- **imgur.com**\n  - `https://imgur.com/a/<id>`\n  - `https://imgur.com/gallery/<id>`\n  - `https://imgur.com/t/<tag>/<id>`\n- **pixiv.net**\n  - `https://www.pixiv.net/artworks/<id>`\n  - `https://www.pixiv.net/<lang>/artworks/<id>`\n\n  - Other sites with the same content as pixiv:\n    - `https://zettai.moe/detail?id=<id>`\n- **reddit.com**\n  - `https://redd.it/<id>`\n  - `https://www.reddit.com/comments/<id>`\n  - `https://www.reddit.com/gallery/<id>`\n  - `https://www.reddit.com/r/<subreddit>/comments/<id>/<any name>`\n- **twitter.com**\n  - `https://(mobile.)twitter.com/<account>/status/<id>`\n  - `https://nitter.net/<account>/status/<id>`\n- **wallhaven.cc**\n  - `https://wallhaven.cc/w/<id>`\n  - `https://whvn.cc/<id>`\n',
    'author': 'Ilia',
    'author_email': 'istudyatuni@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/istudyatuni/artworks-downloader',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from importlib import import_module
from typing import Any, Callable, Coroutine

MODULE = 'art_dl.sites.'


def download(slug: str) -> Callable[[list[str], str], Coroutine[Any, Any, None]]:
	return import_module(MODULE + slug).download


def register(slug: str) -> Callable[[], None]:
	return import_module(MODULE + slug).register



from sys import version_info

if version_info < (3, 10):
	print('Requires python 3.10+')
	quit(1)

# https://github.com/yt-dlp/yt-dlp/blob/master/yt_dlp/__main__.py
import sys

if __package__ is None and not hasattr(sys, 'frozen'):
	# direct call of __main__.py
	import os.path
	path = os.path.realpath(os.path.abspath(__file__))
	sys.path.insert(0, os.path.dirname(os.path.dirname(path)))

import art_dl

if __name__ == '__main__':
	art_dl.main()



# from aiohttp import ClientSession
from asyncio import sleep
from typing import Any, AsyncGenerator

from art_dl.cache import cache
from art_dl.utils.credentials import creds
from art_dl.utils.proxy import ClientSession, ProxyClientSession

from .common import (
	AUTH_LOG_PREFIX,
	BASE_URL,
	CREDS_PATHS,
	REDIRECT_URI,
	SLUG,
	logger,
	make_cache_key,
	progress,
)

API_URL = '/api/v1/oauth2'
# start from 32 instead of 1 to skip small timeouts because
# every time I tested it, we always hit 32+ seconds
# but with 32 at start we can wait in most cases only 32 seconds
# (I don't know why, maybe it not works everytime)
DEFAULT_RATE_LIMIT_TIMEOUT = 32
INVALID_CODE_MSG = 'Incorrect authorization code.'


# TODO: add revoke
# https://www.deviantart.com/developers/authentication
class DAService():
	"""Perform almost all work with auth and API"""

	def __init__(self):
		self.client_id = creds.get(CREDS_PATHS.client_id)
		self.client_secret = creds.get(CREDS_PATHS.client_secret)
		self.code = creds.get(CREDS_PATHS.code)

		self.access_token = creds.get(CREDS_PATHS.access_token)
		self.refresh_token = creds.get(CREDS_PATHS.refresh_token)

	@property
	def _headers(self):
		return {
			'authorization': 'Bearer ' + self.access_token
		}

	def _save_tokens(self):
		creds.delete(CREDS_PATHS.code)
		creds.save(CREDS_PATHS.access_token, self.access_token)
		creds.save(CREDS_PATHS.refresh_token, self.refresh_token)

	async def _ensure_access(self):
		if self.refresh_token is None:
			return await self._fetch_access_token()

		async with ProxyClientSession(BASE_URL) as session:
			async with session.post(
				'/api/v1/oauth2/placebo', params={ 'access_token': self.access_token }
			) as response:
				if (await response.json())['status'] == 'success':
					return

		await self._refresh_token()

	async def _fetch_access_token(self):
		"""Fetch `access_token` using `authorization_code`"""
		if self.code is None:
			logger.warn('authorize app first', prefix=AUTH_LOG_PREFIX)
			quit(1)

		params = {
			'grant_type': 'authorization_code',
			'code': self.code,
			'redirect_uri': REDIRECT_URI,
		}
		await self._authorize(params)

	async def _refresh_token(self):
		"""Refresh `access_token` using `refresh_token`"""
		if self.refresh_token is None:
			raise Exception('No refresh token')

		params = {
			'grant_type': 'refresh_token',
			'refresh_token': self.refresh_token,
		}
		await self._authorize(params)

	async def _authorize(self, add_params):
		""" Common action for _refresh_token and _fetch_access_token """
		params = {
			'client_id': self.client_id,
			'client_secret': self.client_secret,
			**add_params,
		}
		async with ProxyClientSession(BASE_URL) as session:
			async with session.post('/oauth2/token', params=params) as response:
				data = await response.json()
				if response.ok:
					self.access_token = data['access_token']
					self.refresh_token = data['refresh_token']
					self._save_tokens()
				elif data['error_description'] == INVALID_CODE_MSG:
					logger.warn('please authorize again', prefix=AUTH_LOG_PREFIX)
					# or refresh token instead
				elif 'error' in data:
					logger.warn(
						'an error occured during authorization:',
						data['error_description'],
						prefix=AUTH_LOG_PREFIX
					)
					quit(1)

	async def _pager(self, session: ClientSession, method: str, url: str,
						**kwargs) -> AsyncGenerator[Any, None]:
		rate_limit_sec = DEFAULT_RATE_LIMIT_TIMEOUT
		params = {
			**kwargs.pop('params', {}),
			'offset': 0,
			'limit': 24,
			'mature_content': 'true',
		}
		while True:
			async with session.request(method, url, params=params, **kwargs) as response:
				data = await response.json()

				# Rate limit: https://www.deviantart.com/developers/errors
				if response.status == 429:
					if rate_limit_sec > 64 * 10:  # 10 min
						await self._ensure_access()

					u = params['username']
					logger.info(
						f'rate limit in pager ({u}), offset',
						params['offset'],
						'retrying in',
						rate_limit_sec,
						'sec',
						progress=progress
					)
					await sleep(rate_limit_sec)

					rate_limit_sec *= 2
					continue
				elif rate_limit_sec != DEFAULT_RATE_LIMIT_TIMEOUT:
					rate_limit_sec = DEFAULT_RATE_LIMIT_TIMEOUT
					await self._ensure_access()
				elif 'error' in data:
					logger.warn('an error occured during fetching', response.url, progress=progress)
					logger.warn(' ', data['error_description'])
					quit(1)

				response.raise_for_status()

				for result in data['results']:
					yield result

				if data['has_more'] is False:
					break

				params['offset'] = data['next_offset']

	async def list_folders(self, username: str) -> AsyncGenerator[Any, None]:
		await self._ensure_access()

		params = {
			'username': username
		}
		url = f'{API_URL}/gallery/folders'
		async with ProxyClientSession(BASE_URL, headers=self._headers) as session:
			async for folder in self._pager(session, 'GET', url, params=params):
				name = folder['name']
				# i don't know what is this, so just tell about
				if folder['has_subfolders'] is True:
					logger.warn(
						'folder', name, 'has subfolders, but this feature currently not supported'
					)
				yield {
					'id': folder['folderid'],
					'name': name.lower().replace(' ', '-'),
					'pretty_name': name,
				}

	async def list_folder_arts(self, username: str, folder_id: str) -> AsyncGenerator[Any, None]:
		await self._ensure_access()

		params = {
			'username': username
		}
		url = f'{API_URL}/gallery/{folder_id}'
		async with ProxyClientSession(BASE_URL, headers=self._headers) as session:
			async for art in self._pager(session, 'GET', url, params=params):
				if art is not None:
					cache.insert(
						SLUG, make_cache_key(art['author']['username'], art['url']),
						art['deviationid']
					)
				yield art

	async def get_download(self, deviationid: str):
		await self._ensure_access()

		url = f'{API_URL}/deviation/download/{deviationid}'
		async with ProxyClientSession(BASE_URL, headers=self._headers) as session:
			async with session.get(url) as response:
				data = await response.json()
				if 'error' in data:
					logger.warn(
						'error when getting download link:',
						data['error_description'],
						progress=progress
					)
					return
				return data['src']

	async def get_art_info(self, deviationid: str, _rate_limit_sec=DEFAULT_RATE_LIMIT_TIMEOUT):
		await self._ensure_access()

		url = f'{API_URL}/deviation/{deviationid}'
		async with ProxyClientSession(BASE_URL, headers=self._headers) as session:
			async with session.get(url) as response:
				if response.status == 429:
					logger.info(
						'rate limit reached, spleeping for',
						_rate_limit_sec,
						'seconds',
						progress=progress
					)
					await sleep(_rate_limit_sec)
					return await self.get_art_info(deviationid, _rate_limit_sec * 2)
				elif _rate_limit_sec > DEFAULT_RATE_LIMIT_TIMEOUT:
					pass

				data = await response.json()
				if 'error' in data:
					logger.warn(
						'error when getting art info:',
						data['error_description'],
						progress=progress
					)
					return
				return data
